//
//=======================================================================
// Copyright (c) 2004 Kristopher Beevers
//
// Distributed under the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//=======================================================================
//


#include <boost/graph/astar_search.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/random.hpp>
#include <boost/random.hpp>
#include <boost/graph/graphviz.hpp>
#include <sys/time.h>
#include <vector>
#include <list>
#include <iostream>
#include <fstream>
#include <math.h>    // for sqrt

#include <stdio.h>


#include "IvMath/IvVector3.h"
#include "IvMath/IvLineSegment3.h"

using namespace boost;
using namespace std;



// auxiliary types
struct location
{
  float x, y, z;
};
typedef float cost;



template <class Name, class LocMap>
class city_writer {
public:
  city_writer(Name n, LocMap l, float _minx, float _maxx,
              float _miny, float _maxy,
              unsigned int _ptx, unsigned int _pty)
    : name(n), loc(l), minx(_minx), maxx(_maxx), miny(_miny),
      maxy(_maxy), ptx(_ptx), pty(_pty) {}
  template <class Vertex>
  void operator()(ostream& out, const Vertex& v) const {
    float px = 1 - (loc[v].x - minx) / (maxx - minx);
    float py = (loc[v].y - miny) / (maxy - miny);
    out << "[label=\"" << name[v] << "\", pos=\""
        << static_cast<unsigned int>(ptx * px) << ","
        << static_cast<unsigned int>(pty * py)
        << "\", fontsize=\"11\"]";
  }
private:
  Name name;
  LocMap loc;
  float minx, maxx, miny, maxy;
  unsigned int ptx, pty;
};



template <class WeightMap>
class time_writer {
public:
  time_writer(WeightMap w) : wm(w) {}
  template <class Edge>
  void operator()(ostream &out, const Edge& e) const {
    out << "[label=\"" << wm[e] << "\", fontsize=\"11\"]";
  }
private:
  WeightMap wm;
};


// euclidean distance heuristic
template <class Graph, class CostType, class LocMap>
class distance_heuristic : public astar_heuristic<Graph, CostType>
{
public:
  typedef typename graph_traits<Graph>::vertex_descriptor Vertex;
  distance_heuristic(LocMap l, Vertex goal)
    : m_location(l), m_goal(goal) {}
  CostType operator()(Vertex u)
  {
    CostType dx = m_location[m_goal].x - m_location[u].x;
    CostType dy = m_location[m_goal].y - m_location[u].y;
    CostType dz = m_location[m_goal].z - m_location[u].z;
    return ::sqrt(dx*dx + dy*dy + dz*dz);
  }
private:
  LocMap m_location;
  Vertex m_goal;
};


struct found_goal {}; // exception for termination

// visitor that terminates when we find the goal
template <class Vertex>
class astar_goal_visitor : public boost::default_astar_visitor
{
public:
  astar_goal_visitor(Vertex goal) : m_goal(goal) {}
  template <class Graph>
  void examine_vertex(Vertex u, Graph& g) {
    if(u == m_goal)
      throw found_goal();
  }
private:
  Vertex m_goal;
};


typedef adjacency_list<listS, vecS, undirectedS, no_property, property<edge_weight_t, cost> > mygraph_t;
typedef mygraph_t::edge_descriptor edge_descriptor;
typedef property_map<mygraph_t, edge_weight_t>::type WeightMap;

/*
int GRAPH_edge_add(mygraph_t &g, int edge_a, int edge_b)
{
	WeightMap weightmap = get(edge_weight, g);
	
	edge_descriptor e;
	bool inserted;
    boost::tie(e, inserted) = add_edge(
		edge_a,
		edge_b,
		g
	);
	// at first dont care about the weights!
    // weightmap[e] = weights[j]; 
	weightmap[e] = 1;
	
	return 0; // todo: return handle to delete the edge
}
*/

class cWrapperGraph
{
	typedef std::pair<int, int> edge;
	typedef mygraph_t::vertex_descriptor vertex;

	public:
		// dynamic data (the graph is builded from)
		std::list<edge> theEdges;
		std::list<location> theVertices;
		
		// data of the current graph (deleted/newed every buildGraph())
		mygraph_t *theGraph;
		location *locations_;
		
		cWrapperGraph()
		{
			theGraph = NULL;
		}
		
		int getNearestWaypoints(float *pos, int *out_wp1, int *out_wp2)
		{
			//printf("pos=%.2f %.2f %.2f\n", pos[0], pos[1], pos[2]);

						
			edge bestEdge;
			float bestDistance;
			std::list<edge>::iterator edge_i;
			int first = 1;
			for (edge_i=theEdges.begin(); edge_i!=theEdges.end(); edge_i++)
			{
			
				edge tmp = *edge_i;
				
				location a = locations_[tmp.first];
				location b = locations_[tmp.second];
			
				/*printf("edge(%d,%d): (%.2f %.2f %.2f)->(%.2f %.2f %.2f)\n", tmp.first, tmp.second,
					a.x, a.y, a.z,
					b.x, b.y, b.z
				);*/
				
				//printf("30.01.2012 IV lib testing program\n");
				
				IvVector3 from(a.x, a.y, a.z);
				IvVector3 to(b.x, b.y, b.z);
				
				//printf("from.x = %.2f\n", from.GetX());
				
				IvLineSegment3 line(from, to);
				
				float bla1, bla2;
				
				bla1 = DistanceSquared(line, IvVector3(pos[0], pos[1], pos[2]), bla2);
				
				// init the data with the first element
				if (first)
				{
					bestEdge = *edge_i;
					bestDistance = bla1;
					first = 0;
					continue;
				}
				
				if (bla1 < bestDistance)
				{
					bestEdge = *edge_i;
					bestDistance = bla1;
				}
				
				//printf("distance to linesegment = %.2f\n", bla1);				
			}
			
			*out_wp1 = bestEdge.first;
			*out_wp2 = bestEdge.second;
			//printf("best edge:(%d,%d) distance:%.2f\n", bestEdge.first, bestEdge.second, bestDistance);
			
			return 1;
		}
		
		int addEdge(int edgeA, int edgeB)
		{
			// DONT ADD IT DIRECTLY TO THE GRAPH
			// FIRST SaVE THEM IN theEdges
			// THEN PUT THEM IN GRAPH WHEN graph->buildGraph()
			/*WeightMap weightmap = get(edge_weight, *theGraph);

			edge_descriptor e;
			bool inserted;
			boost::tie(e, inserted) = add_edge(
				edgeA,
				edgeB,
				*theGraph
			);
			// at first dont care about the weights!
			// weightmap[e] = weights[j]; 
			weightmap[e] = 1;
			*/
			
			theEdges.push_back(edge(edgeA,edgeB));
			
			return 1; // todo: return handle to delete the edge
		}
		int addEdgeToGraph(edge tmp)
		{
			WeightMap weightmap = get(edge_weight, *theGraph);

			edge_descriptor e;
			bool inserted;
			boost::tie(e, inserted) = add_edge(
				tmp.first,
				tmp.second,
				*theGraph
			);
			// at first dont care about the weights!
			// weightmap[e] = weights[j]; 
			
			float dx = locations_[tmp.first].x - locations_[tmp.second].x;
			float dy = locations_[tmp.first].y - locations_[tmp.second].y;
			float dz = locations_[tmp.first].z - locations_[tmp.second].z;
			float theCost = ::sqrt(dx*dx + dy*dy + dz*dz);
			
			//printf("weight=%.2f\n", theCost);
			weightmap[e] = theCost;
			
			return 1; // todo: return handle to delete the edge?!?! not anymore
		}
		
		int addVertex(location tmp)
		{
			theVertices.push_back(tmp);
		}
		
		int buildGraph()
		{
			if (theGraph != NULL)
			{
				delete theGraph;
				free(locations_);
			}
			
			theGraph = new mygraph_t(theVertices.size());
			
			locations_ = (location *) malloc(theVertices.size() * sizeof(location));
			std::list<location>::iterator vertex_i;
			int i = 0;
			for (vertex_i=theVertices.begin(); vertex_i!=theVertices.end(); vertex_i++)
			{
				locations_[i].x = (*vertex_i).x;
				locations_[i].y = (*vertex_i).y;
				locations_[i].z = (*vertex_i).z;
				i++;
			}
			
			std::list<edge>::iterator edge_i;
			for (edge_i=theEdges.begin(); edge_i!=theEdges.end(); edge_i++)
				addEdgeToGraph(*edge_i);
		}
		
		int doAStar(vertex start, vertex goal, std::list<vertex> &ret, float *outCost)
		{
			// specify some types
			// typedef adjacency_list<listS, vecS, undirectedS, no_property, property<edge_weight_t, cost> > mygraph_t;
			//typedef property_map<mygraph_t, edge_weight_t>::type WeightMap;
			//typedef mygraph_t::vertex_descriptor vertex;
			//typedef mygraph_t::edge_descriptor edge_descriptor;
			typedef mygraph_t::vertex_iterator vertex_iterator;

			const char *name[] = {"A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P"};

			// num = 3 (for 0 1 2)
			// goal = 3 (0 1 2 3 = 4) -> error to prevent
			//printf("num_vertices=%d\n", num_vertices(*theGraph));
			if (num_vertices(*theGraph) <= start)
				return 0;
			if (num_vertices(*theGraph) <= goal)
				return 0;
			
			/*
			// between the edges! shouldnt that be calculated from the location!??!?
			cost weights[] = { // estimated travel time (mins)
			96, 134, 143, 65, 115, 133, 117, 116, 74, 56,
			84, 73, 69, 70, 116, 147, 173, 183, 74, 71, 124
			};
			cost *weights = (cost *)malloc(num_edges * sizeof(cost));
			for (int i=0; i<num_edges; i++)
			weights[i] =  1; // maybe the distance here? but dont know how to read out the edges -.-

			*/

			/*
			WeightMap weightmap = get(edge_weight, g);
			for(int j=0; j<num_edges; j++) {
			edge_descriptor e;
			bool inserted;
			boost::tie(e, inserted) = add_edge(
			edge_array[j].first,
			edge_array[j].second,
			g
			);
			// at first dont care about the weights!
			// weightmap[e] = weights[j]; 
			weightmap[e] = 1;

			//printf("edge[%d].inserted = %d (%s->%s)\n", j, inserted, name[edge_array[j].first], name[edge_array[j].second]);
			}*/

			/*
			ofstream dotfile;
			dotfile.open("test-astar-cities.dot");
			write_graphviz(
			dotfile,
			g,
			city_writer<const char **, location*>(name, locations, 73.46, 78.86, 40.67, 44.93,480, 400),
			time_writer<WeightMap>(weightmap)
			);
			*/

			std::vector<mygraph_t::vertex_descriptor> p(num_vertices(*theGraph));
			std::vector<cost> d(num_vertices(*theGraph));
			try {
				// call astar named parameter interface
				astar_search(
					*theGraph,
					start,
					distance_heuristic<mygraph_t, cost, location*>(locations_, goal),
					predecessor_map(&p[0]).distance_map(&d[0]).visitor(astar_goal_visitor<vertex>(goal))
				);
			} catch(found_goal fg) {
				// p = parent?!
				for(vertex v=goal;; v=p[v]) {
					//printf("push_front(%d = '%s')\n", v, name[v]);
					ret.push_front(v);
					if(p[v] == v)
						break;
				}
				//cout << endl << "Total travel time: " << d[goal] << endl;
				
				// ups, theres just a problem with the weights-calculation!
				*outCost = d[goal];
				
				//printf("cost bla=%.2f\n", d[goal]);
				//printf("cost foo=%.2f\n", (float)d[goal]);
				//printf("cost foo=%d\n", (int)d[goal]);
				
				// FUCK YOU TYPEDEF FROM FLOAT TO COST -.-
				// CANT EVEN CONVERT IT NORMALLY -.-
				/*ostringstream buf;
				buf << d[goal];
				//printf("%s", buf.str());
				*outCost = atof(buf.str().c_str());*/
			}
		}
		~cWrapperGraph()
		{
			printf("DEconstructed!\n");
		}
};

extern "C"
{

int GRAPH_new()
{
	cWrapperGraph *graph;
	graph = new cWrapperGraph;
	printf("astar.cpp: graph=%d\n", graph);
	return (int) graph;
}

int GRAPH_add_edge(int graph, int first, int second)
{
	cWrapperGraph *tmp = (cWrapperGraph *)graph;
	tmp->addEdge(first, second);
	return 1;
}
int GRAPH_add_vertex(int graph, float x, float y, float z)
{
	cWrapperGraph *tmp = (cWrapperGraph *)graph;
	struct location bla = {x, y, z};
	//printf("vertex: %.2f %.2f %.2f\n", bla.x, bla.y, bla.z);
	tmp->addVertex(bla);
	return 1;
}
int GRAPH_delete(int graph)
{
	cWrapperGraph *tmp = (cWrapperGraph *)graph;
	delete tmp;
}
int GRAPH_build(int graph)
{
	cWrapperGraph *tmp = (cWrapperGraph *)graph;
	tmp->buildGraph();
	return 1;
}
int *GRAPH_do_astar(int graph, int from, int to, int **outRet, int *outLen, float *outCost)
{
	typedef mygraph_t::vertex_descriptor vertex;
	cWrapperGraph *tmp = (cWrapperGraph *)graph;
	std::list<vertex> ret;
	vertex a = from;
	vertex b = to;
	
	float cost;
	
	//printf("before\n");
	tmp->doAStar(a, b, ret, &cost);
	//printf("after\n");
	
	//printf("cost 1: %.2f\n", cost);
	
	//if ()
	
	int *ptr = (int *) malloc(ret.size() * sizeof(int));
	int i = 0;
	if (ptr == NULL)
	{
		printf("ptr == NULL!\n");
		*outRet = NULL;
		*outLen = NULL;
		return NULL;
	}
	
	for(std::list<vertex>::iterator spi=ret.begin(); spi!=ret.end(); spi++)
	{
		ptr[i] = *spi;
		i++;
	}
	*outRet = ptr;
	*outLen = ret.size();
	
	*outCost = cost;
	
	return ptr;
}


/*
	graph = pointer to graph-class
	pos = position of player or zombie
	out_wp1 = id of waypoint
	out_wp2 = id of waypoint other side
	out_pos = position between the two waypoints 
*/
int GRAPH_getNearestEdge(int graph, float *pos, int *out_wp1, int *out_wp2, float **out_pos)
{
	cWrapperGraph *tmp = (cWrapperGraph *)graph;
	//struct location bla = {x, y, z};
	//printf("vertex: %.2f %.2f %.2f\n", bla.x, bla.y, bla.z);
	tmp->getNearestWaypoints(pos, out_wp1, out_wp2);
	//printf("graph->getNearestWaypoints() -> %d %d\n", *out_wp1, *out_wp2);
	return 1;
}

	int MATH_nearest_point_on_linesegment(float *from, float *to, float *toPoint, float *out_point)
	{
		/*
		out_point[0] = (from[0] + to[0]) / 2;
		out_point[1] = (from[1] + to[1]) / 2;
		out_point[2] = (from[2] + to[2]) / 2;
		*/
		
		IvVector3 from_(from[0], from[1], from[2]);
		IvVector3 to_(to[0], to[1], to[2]);
				
				//printf("from.x = %.2f\n", from.GetX());
				
		IvLineSegment3 line(from_, to_);
				
		IvVector3 ret, toPoint_(toPoint[0], toPoint[1], toPoint[2]);
				
		ret = line.ClosestPoint(toPoint_);
		
		out_point[0] = ret.GetX();
		out_point[1] = ret.GetY();
		out_point[2] = ret.GetZ();
		
		//printf("(%.2f, %.2f, %.2f)\n", out_point[0], out_point[1], out_point[2]);
		
		return 1;
	}

}

/*
root@vps35166:/home/cod2_1.3_0/astar# g++ -m32  astar.cpp -O3 -Wno-deprecated -DWITHMAIN
root@vps35166:/home/cod2_1.3_0/astar# ./a.out



root@vps35166:/home/cod2_1.3_0/astar# g++ -m32  astar.cpp -O3 -Wno-deprecated -DWITHMAIN -I../iv -I../iv/IvUtility -L../iv/ -liv
/usr/bin/ld: skipping incompatible ../iv//libiv.a when searching for -liv
/usr/bin/ld: cannot find -liv
collect2: ld returned 1 exit status
root@vps35166:/home/cod2_1.3_0/astar#


*/

#ifdef WITHMAIN
int main()
{
	typedef std::pair<int, int> edge;
	typedef mygraph_t::vertex_descriptor vertex;

	enum nodes {
	A, B, C, D, E, F, G, H, I, J, K, L, M, N, O, P, NUMBER_OF_VERTEXES
	};

	// redundant edges, for easy input
	edge edge_array[] = {
	edge(A,B),	edge(A,I),

	edge(B,A),
	edge(B,C),
	edge(B,E),
	edge(B,F),

	edge(C,B),
	edge(C,D),
	edge(C,E),
	edge(C,F),

	edge(D,C),
	edge(D,G),
	edge(D,H),

	edge(E,B),
	edge(E,C),
	edge(E,F),
	edge(E,J),

	edge(F,B),
	edge(F,C),
	edge(F,E),
	edge(F,K),

	edge(G,D),
	edge(G,H),
	edge(G,L),

	edge(H,D),
	edge(H,G),
	edge(H,P),

	edge(I,A),
	edge(I,J),
	edge(I,M),

	edge(J,E),
	edge(J,I),
	edge(J,M),

	edge(K,F),
	edge(K,L),
	edge(K,O),
	edge(K,N),

	edge(L,G),
	edge(L,K),
	edge(L,N),
	edge(L,O),

	edge(M,I),
	edge(M,J),
	edge(M,N),

	edge(N,M),
	edge(N,K),
	edge(N,L),
	edge(N,O),

	edge(O,N),
	edge(O,K),
	edge(O,L),
	edge(O,P),

	edge(P,O),
	edge(P,H)
	};
	unsigned int num_edges = sizeof(edge_array) / sizeof(edge);
	location locations[] = {
		{76,48,0}, // A
		{325,44,0}, // B
		{578,43,0}, // C
		{803,41,0}, // D
		{326,137,0}, // E
		{583,142,0}, // F
		{804,148,0}, // G
		{1030,149,0}, // H
		{79,238,0}, // I
		{322,243,0}, // J
		{583,232,0}, // K
		{802,229,0}, // L
		{319,347,1000}, // M
		{583,343,0}, // N
		{800,351,0}, // O
		{1032,349,0} // P
	};
	const char *name[] = {"A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P"};
	int wrapper = GRAPH_new();

	for (int i=0; i<num_edges; i++)
		GRAPH_add_edge(wrapper, edge_array[i].first, edge_array[i].second);
	for (int i=0; i<NUMBER_OF_VERTEXES; i++)
		GRAPH_add_vertex(wrapper, locations[i].x, locations[i].y, locations[i].z);

	// set up malloc()-arrays for fast access
	GRAPH_build(wrapper);
	
	
	//int GRAPH_getNearestEdge(int graph, float *pos, int *out_wp1, int *out_wp2, float **out_pos)
	float pos[3] = {150,40,0};
	int wp1, wp2;
	GRAPH_getNearestEdge(wrapper, pos, &wp1, &wp2, NULL);
		

	// top-left to bottom-right
	vertex start = 0; // A
	vertex goal = 15; // P

	cout << "Start vertex: " << name[start] << endl;
	cout << "Goal vertex: " << name[goal] << endl;

	//printf("num_vertices=%d\n", num_vertices(*theGraph)/*, num_edges(g)*/);

	std::list<vertex> ret;
	
	int *outRet;
	int outLen;
	float cost;
	
	GRAPH_do_astar(wrapper, start, goal, &outRet, &outLen, &cost);
	for (int i=0; i<outLen; i++)
		printf("%s ", name[outRet[i]]);
/*	wrapper->doAStar(start, goal, ret);

	printf("Shortest path from %s to %s: ", name[start], name[goal]);
	for(std::list<vertex>::iterator spi=ret.begin(); spi!=ret.end(); spi++)
		printf("%s ", name[*spi]);
*/
		
	GRAPH_delete(wrapper);

	return 0;
}
#endif 

#if 0
int main1()
{
	typedef std::pair<int, int> edge;
	typedef mygraph_t::vertex_descriptor vertex;

	enum nodes {
	A, B, C, D, E, F, G, H, I, J, K, L, M, N, O, P, NUMBER_OF_VERTEXES
	};

	// redundant edges, for easy input
	edge edge_array[] = {
	edge(A,B),	edge(A,I),

	edge(B,A),
	edge(B,C),
	edge(B,E),
	edge(B,F),

	edge(C,B),
	edge(C,D),
	edge(C,E),
	edge(C,F),

	edge(D,C),
	edge(D,G),
	edge(D,H),

	edge(E,B),
	edge(E,C),
	edge(E,F),
	edge(E,J),

	edge(F,B),
	edge(F,C),
	edge(F,E),
	edge(F,K),

	edge(G,D),
	edge(G,H),
	edge(G,L),

	edge(H,D),
	edge(H,G),
	edge(H,P),

	edge(I,A),
	edge(I,J),
	edge(I,M),

	edge(J,E),
	edge(J,I),
	edge(J,M),

	edge(K,F),
	edge(K,L),
	edge(K,O),
	edge(K,N),

	edge(L,G),
	edge(L,K),
	edge(L,N),
	edge(L,O),

	edge(M,I),
	edge(M,J),
	edge(M,N),

	edge(N,M),
	edge(N,K),
	edge(N,L),
	edge(N,O),

	edge(O,N),
	edge(O,K),
	edge(O,L),
	edge(O,P),

	edge(P,O),
	edge(P,H)
	};
	unsigned int num_edges = sizeof(edge_array) / sizeof(edge);
	location locations[] = {
		{76,48,0}, // A
		{325,44,0}, // B
		{578,43,0}, // C
		{803,41,0}, // D
		{326,137,0}, // E
		{583,142,0}, // F
		{804,148,0}, // G
		{1030,149,0}, // H
		{79,238,0}, // I
		{322,243,0}, // J
		{583,232,0}, // K
		{802,229,0}, // L
		{319,347,0}, // M
		{583,343,0}, // N
		{800,351,0}, // O
		{1032,349,0} // P
	};
	const char *name[] = {"A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P"};
	cWrapperGraph *wrapper = new cWrapperGraph;

	for (int i=0; i<num_edges; i++)
		wrapper->addEdge(edge_array[i].first, edge_array[i].second);
	for (int i=0; i<NUMBER_OF_VERTEXES; i++)
		wrapper->addVertex(locations[i]);

	wrapper->buildGraph();

	// top-left to bottom-right
	vertex start = 0; // A
	vertex goal = 15; // P

	cout << "Start vertex: " << name[start] << endl;
	cout << "Goal vertex: " << name[goal] << endl;

	//printf("num_vertices=%d\n", num_vertices(*theGraph)/*, num_edges(g)*/);

	std::list<vertex> ret;
	wrapper->doAStar(start, goal, ret);

	printf("Shortest path from %s to %s: ", name[start], name[goal]);
	for(std::list<vertex>::iterator spi=ret.begin(); spi!=ret.end(); spi++)
		printf("%s ", name[*spi]);

	delete wrapper;

	return 0;
}
#endif