//
//=======================================================================
// Copyright (c) 2004 Kristopher Beevers
//
// Distributed under the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//=======================================================================
//


#include <boost/graph/astar_search.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/random.hpp>
#include <boost/random.hpp>
#include <boost/graph/graphviz.hpp>
#include <sys/time.h>
#include <vector>
#include <list>
#include <iostream>
#include <fstream>
#include <math.h>    // for sqrt

#include <stdio.h>

using namespace boost;
using namespace std;



// auxiliary types
struct location
{
  float x, y, z;
};
typedef float cost;



template <class Name, class LocMap>
class city_writer {
public:
  city_writer(Name n, LocMap l, float _minx, float _maxx,
              float _miny, float _maxy,
              unsigned int _ptx, unsigned int _pty)
    : name(n), loc(l), minx(_minx), maxx(_maxx), miny(_miny),
      maxy(_maxy), ptx(_ptx), pty(_pty) {}
  template <class Vertex>
  void operator()(ostream& out, const Vertex& v) const {
    float px = 1 - (loc[v].x - minx) / (maxx - minx);
    float py = (loc[v].y - miny) / (maxy - miny);
    out << "[label=\"" << name[v] << "\", pos=\""
        << static_cast<unsigned int>(ptx * px) << ","
        << static_cast<unsigned int>(pty * py)
        << "\", fontsize=\"11\"]";
  }
private:
  Name name;
  LocMap loc;
  float minx, maxx, miny, maxy;
  unsigned int ptx, pty;
};



template <class WeightMap>
class time_writer {
public:
  time_writer(WeightMap w) : wm(w) {}
  template <class Edge>
  void operator()(ostream &out, const Edge& e) const {
    out << "[label=\"" << wm[e] << "\", fontsize=\"11\"]";
  }
private:
  WeightMap wm;
};


// euclidean distance heuristic
template <class Graph, class CostType, class LocMap>
class distance_heuristic : public astar_heuristic<Graph, CostType>
{
public:
  typedef typename graph_traits<Graph>::vertex_descriptor Vertex;
  distance_heuristic(LocMap l, Vertex goal)
    : m_location(l), m_goal(goal) {}
  CostType operator()(Vertex u)
  {
    CostType dx = m_location[m_goal].x - m_location[u].x;
    CostType dy = m_location[m_goal].y - m_location[u].y;
    CostType dz = m_location[m_goal].z - m_location[u].z;
    return ::sqrt(dx*dx + dy*dy + dz*dz);
  }
private:
  LocMap m_location;
  Vertex m_goal;
};


struct found_goal {}; // exception for termination

// visitor that terminates when we find the goal
template <class Vertex>
class astar_goal_visitor : public boost::default_astar_visitor
{
public:
  astar_goal_visitor(Vertex goal) : m_goal(goal) {}
  template <class Graph>
  void examine_vertex(Vertex u, Graph& g) {
    if(u == m_goal)
      throw found_goal();
  }
private:
  Vertex m_goal;
};


typedef adjacency_list<listS, vecS, undirectedS, no_property, property<edge_weight_t, cost> > mygraph_t;
typedef mygraph_t::edge_descriptor edge_descriptor;
typedef property_map<mygraph_t, edge_weight_t>::type WeightMap;

int GRAPH_edge_add(mygraph_t &g, int edge_a, int edge_b)
{
	WeightMap weightmap = get(edge_weight, g);
	
	edge_descriptor e;
	bool inserted;
    boost::tie(e, inserted) = add_edge(
		edge_a,
		edge_b,
		g
	);
	// at first dont care about the weights!
    // weightmap[e] = weights[j]; 
	weightmap[e] = 1;
	
	return 0; // todo: return handle to delete the edge
}

class cWrapperGraph
{
			typedef std::pair<int, int> edge;
			typedef mygraph_t::vertex_descriptor vertex;
	public:
		// dynamic data (the graph is builded from)
		std::list<edge> theEdges;
		std::list<location> theVertices;
		
		// data of the current graph (deleted/newed every buildGraph())
		mygraph_t *theGraph;
		location *locations_;
		
		cWrapperGraph()
		{
			theGraph = NULL;
		}
		
		int addEdge(int edgeA, int edgeB)
		{
			// DONT ADD IT DIRECTLY TO THE GRAPH
			// FIRST SaVE THEM IN theEdges
			// THEN PUT THEM IN GRAPH WHEN graph->buildGraph()
			/*WeightMap weightmap = get(edge_weight, *theGraph);

			edge_descriptor e;
			bool inserted;
			boost::tie(e, inserted) = add_edge(
				edgeA,
				edgeB,
				*theGraph
			);
			// at first dont care about the weights!
			// weightmap[e] = weights[j]; 
			weightmap[e] = 1;
			*/
			
			theEdges.push_back(edge(edgeA,edgeB));
			
			return 1; // todo: return handle to delete the edge
		}
		int addEdgeToGraph(edge tmp)
		{
			WeightMap weightmap = get(edge_weight, *theGraph);

			edge_descriptor e;
			bool inserted;
			boost::tie(e, inserted) = add_edge(
				tmp.first,
				tmp.second,
				*theGraph
			);
			// at first dont care about the weights!
			// weightmap[e] = weights[j]; 
			weightmap[e] = 1;
			
			return 1; // todo: return handle to delete the edge?!?! not anymore
		}
		
		int addVertex(location tmp)
		{
			theVertices.push_back(tmp);
		}
		
		int buildGraph()
		{
			if (theGraph != NULL)
			{
				delete theGraph;
				free(locations_);
			}
			locations_ = (location *) malloc(theVertices.size() * sizeof(location));
			theGraph = new mygraph_t(theVertices.size());
			
			std::list<edge>::iterator edge_i;
			for (edge_i=theEdges.begin(); edge_i!=theEdges.end(); edge_i++)
				addEdgeToGraph(*edge_i);
		}
		
		int doAStar(vertex start, vertex goal, std::list<vertex> &ret)
		{
			// specify some types
			// typedef adjacency_list<listS, vecS, undirectedS, no_property, property<edge_weight_t, cost> > mygraph_t;
			//typedef property_map<mygraph_t, edge_weight_t>::type WeightMap;
			//typedef mygraph_t::vertex_descriptor vertex;
			//typedef mygraph_t::edge_descriptor edge_descriptor;
			typedef mygraph_t::vertex_iterator vertex_iterator;

			const char *name[] = {"A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P"};

			/*
			// between the edges! shouldnt that be calculated from the location!??!?
			cost weights[] = { // estimated travel time (mins)
			96, 134, 143, 65, 115, 133, 117, 116, 74, 56,
			84, 73, 69, 70, 116, 147, 173, 183, 74, 71, 124
			};
			cost *weights = (cost *)malloc(num_edges * sizeof(cost));
			for (int i=0; i<num_edges; i++)
			weights[i] =  1; // maybe the distance here? but dont know how to read out the edges -.-

			*/

			/*
			WeightMap weightmap = get(edge_weight, g);
			for(int j=0; j<num_edges; j++) {
			edge_descriptor e;
			bool inserted;
			boost::tie(e, inserted) = add_edge(
			edge_array[j].first,
			edge_array[j].second,
			g
			);
			// at first dont care about the weights!
			// weightmap[e] = weights[j]; 
			weightmap[e] = 1;

			//printf("edge[%d].inserted = %d (%s->%s)\n", j, inserted, name[edge_array[j].first], name[edge_array[j].second]);
			}*/

			/*
			ofstream dotfile;
			dotfile.open("test-astar-cities.dot");
			write_graphviz(
			dotfile,
			g,
			city_writer<const char **, location*>(name, locations, 73.46, 78.86, 40.67, 44.93,480, 400),
			time_writer<WeightMap>(weightmap)
			);
			*/

			std::vector<mygraph_t::vertex_descriptor> p(num_vertices(*theGraph));
			std::vector<cost> d(num_vertices(*theGraph));
			try {
				// call astar named parameter interface
				astar_search(
					*theGraph,
					start,
					distance_heuristic<mygraph_t, cost, location*>(locations_, goal),
					predecessor_map(&p[0]).distance_map(&d[0]).visitor(astar_goal_visitor<vertex>(goal))
				);
			} catch(found_goal fg) {
				// p = parent?!
				for(vertex v=goal;; v=p[v]) {
					//printf("push_front(%d = '%s')\n", v, name[v]);
					ret.push_front(v);
					if(p[v] == v)
						break;
				}
				//cout << endl << "Total travel time: " << d[goal] << endl;
			}
		}
		~cWrapperGraph()
		{
			printf("DEconstructed!\n");
		}
};

int main()
{
	typedef std::pair<int, int> edge;
			typedef mygraph_t::vertex_descriptor vertex;

	enum nodes {
	A, B, C, D, E, F, G, H, I, J, K, L, M, N, O, P, NUMBER_OF_VERTEXES
	};
			
	// redundant edges, for easy input
	edge edge_array[] = {
	edge(A,B),
	edge(A,I),

	edge(B,A),
	edge(B,C),
	edge(B,E),
	edge(B,F),

	edge(C,B),
	edge(C,D),
	edge(C,E),
	edge(C,F),

	edge(D,C),
	edge(D,G),
	edge(D,H),

	edge(E,B),
	edge(E,C),
	edge(E,F),
	edge(E,J),

	edge(F,B),
	edge(F,C),
	edge(F,E),
	edge(F,K),

	edge(G,D),
	edge(G,H),
	edge(G,L),

	edge(H,D),
	edge(H,G),
	edge(H,P),

	edge(I,A),
	edge(I,J),
	edge(I,M),

	edge(J,E),
	edge(J,I),
	edge(J,M),

	edge(K,F),
	edge(K,L),
	edge(K,O),
	edge(K,N),

	edge(L,G),
	edge(L,K),
	edge(L,N),
	edge(L,O),

	edge(M,I),
	edge(M,J),
	edge(M,N),

	edge(N,M),
	edge(N,K),
	edge(N,L),
	edge(N,O),

	edge(O,N),
	edge(O,K),
	edge(O,L),
	edge(O,P),

	edge(P,O),
	edge(P,H)
	};
	unsigned int num_edges = sizeof(edge_array) / sizeof(edge);
	location locations[] = {
	{76,48,0}, {325,44,0}, {578,43,0}, {803,41,0}, {326,137,0}, {583,142,0}, {804,148,0},
	{1030,149,0}, {79,238,0}, {322,243,0}, {583,232,0}, {802,229,0}, {319,347,0}, {583,343,0},
	{800,351,0}, {1032,349,0}
	};
	const char *name[] = {"A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P"};
	cWrapperGraph *wrapper = new cWrapperGraph;
	
	for (int i=0; i<num_edges; i++)
		wrapper->addEdge(edge_array[i].first, edge_array[i].second);
	for (int i=0; i<NUMBER_OF_VERTEXES; i++)
		wrapper->addVertex(locations[i]);
	
	wrapper->buildGraph();
	
		vertex start, goal;


			// top-left to bottom-right
			start = 0; // A
			goal = 15; // P

			cout << "Start vertex: " << name[start] << endl;
			cout << "Goal vertex: " << name[goal] << endl;

			//printf("num_vertices=%d\n", num_vertices(*theGraph)/*, num_edges(g)*/);
	
	std::list<vertex> ret;
	wrapper->doAStar(start, goal, ret);
				std::list<vertex>::iterator spi;
				
				
			printf("Shortest path from %s to %s: ", name[start], name[goal]);
				
			for(spi=ret.begin(); spi!=ret.end(); spi++)
			printf(" ----%s", name[*spi]);
	delete wrapper;

  return 0;
  
}