/*! Sunrise Data Dictionary Library
 *
 *  @brief a library for hashtable storage of arbitrary data objects
 *  with built-in reference counting and guaranteed order iteration.
 *
 *  @file dd_settings.pp
 *  post-process compile-time settings
 *
 *  This file can be edited by library users for customisation.
 *  The library needs to be rebuild for changes to take effect.
 *
 *  Created by Sunrise Telephone Systems Ltd.
 *
 *  This file ("dd_settings.pp") is hereby placed into the public domain.
 *
 */

#ifndef _DD_SETTINGS_CHECKED
#define _DD_SETTINGS_CHECKED

#include "dd_settings.h"

// --------------------------------------------------------------------------
// Range check for DD_MAXIMUM_KEY_LENGTH,  must be > 0
// --------------------------------------------------------------------------

#ifndef DD_MAXIMUM_KEY_LENGTH
#define DD_MAXIMUM_KEY_LENGTH 128
#endif

#if (DD_MAXIMUM_KEY_LENGTH < 1)
#undef DD_MAXIMUM_KEY_LENGTH
#define DD_MAXIMUM_KEY_LENGTH 128
#endif

// --------------------------------------------------------------------------
// Range check for DD_SIGNIFICANT_CHARS,  must be > 0 and <= max key length
// --------------------------------------------------------------------------

#ifndef DD_SIGNIFICANT_CHARS
#define DD_SIGNIFICANT_CHARS 32
#endif

#if ((DD_SIGNIFICANT_CHARS < 1) || (DD_SIGNIFICANT_CHARS > DD_MAXIMUM_KEY_LENGTH))
#undef DD_SIGNIFICANT_CHARS
#define DD_SIGNIFICANT_CHARS 32
#endif

// --------------------------------------------------------------------------
// Value check for DD_SDBM, DD_DJB, DD_FNV, and DD_HSIEH
// --------------------------------------------------------------------------

#ifndef DD_SDBM
#define DD_SDBM 1
#endif

#if (DD_SDBM != 1)
#undef DD_SDBM
#define DD_SDBM 1
#endif

#ifndef DD_DJB
#define DD_DJB 2
#endif

#if (DD_DJB != 2)
#undef DD_DJB
#define DD_DJB 2
#endif

#ifndef DD_ELF
#define DD_ELF 3
#endif

#if (DD_ELF != 3)
#undef DD_ELF
#define DD_ELF 3
#endif

#ifndef DD_FNVM
#define DD_FNVM 4
#endif

#if (DD_FNVM != 4)
#undef DD_FNVM
#define DD_FNVM 4
#endif

#ifndef DD_HSIEH
#define DD_HSIEH 5
#endif

#if (DD_HSIEH != 5)
#undef DD_HSIEH
#define DD_HSIEH 5
#endif

// --------------------------------------------------------------------------
// Range check for DD_DEFAULT_HASH_ALGORITHM
// --------------------------------------------------------------------------
// must be any of DD_SDBM, DD_DJB, DD_ELF, DD_FNVM or DD_HSIEH

#ifndef DD_DEFAULT_HASH_ALGORITHM
#define DD_DEFAULT_HASH_ALGORITHM DD_SDBM
#endif

#if ((DD_DEFAULT_HASH_ALGORITHM != DD_SDBM) && \
     (DD_DEFAULT_HASH_ALGORITHM != DD_DJB) && \
     (DD_DEFAULT_HASH_ALGORITHM != DD_ELF) && \
     (DD_DEFAULT_HASH_ALGORITHM != DD_FNVM) && \
	 (DD_DEFAULT_HASH_ALGORITHM != DD_HSIEH))
#undef DD_DEFAULT_HASH_ALGORITHM
#define DD_DEFAULT_HASH_ALGORITHM DD_SDBM
#endif

// --------------------------------------------------------------------------
// Range check for DD_CASE_SENSITIVE_HASHING,  must be true or false
// --------------------------------------------------------------------------

#ifndef DD_CASE_SENSITIVE_HASHING
#define DD_CASE_SENSITIVE_HASHING true
#endif

#if ((DD_CASE_SENSITIVE_HASHING != true) && (DD_CASE_SENSITIVE_HASHING != false))
#undef DD_CASE_SENSITIVE_HASHING
#define DD_CASE_SENSITIVE_HASHING true
#endif

// --------------------------------------------------------------------------
// Define DD_DEFAULT_HASH_FUNCTION
// --------------------------------------------------------------------------

#if ((DD_DEFAULT_HASH_ALGORITHM == DD_SDBM) && (DD_CASE_SENSITIVE_HASHING == true))
#define DD_DEFAULT_HASH_FUNCTION dd_SDBM_hash_string
#elif ((DD_DEFAULT_HASH_ALGORITHM == DD_SDBM) && (DD_CASE_SENSITIVE_HASHING == false))
#define DD_DEFAULT_HASH_FUNCTION dd_SDBM_hash_string_toupper
#elif ((DD_DEFAULT_HASH_ALGORITHM == DD_DJB) && (DD_CASE_SENSITIVE_HASHING == true))
#define DD_DEFAULT_HASH_FUNCTION dd_DJB_hash_string
#elif ((DD_DEFAULT_HASH_ALGORITHM == DD_DJB) && (DD_CASE_SENSITIVE_HASHING == false))
#define DD_DEFAULT_HASH_FUNCTION dd_DJB_hash_string_toupper
#elif ((DD_DEFAULT_HASH_ALGORITHM == DD_ELF) && (DD_CASE_SENSITIVE_HASHING == true))
#define DD_DEFAULT_HASH_FUNCTION dd_ELF_hash_string
#elif ((DD_DEFAULT_HASH_ALGORITHM == DD_ELF) && (DD_CASE_SENSITIVE_HASHING == false))
#define DD_DEFAULT_HASH_FUNCTION dd_ELF_hash_string_toupper
#elif ((DD_DEFAULT_HASH_ALGORITHM == DD_FNVM) && (DD_CASE_SENSITIVE_HASHING == true))
#define DD_DEFAULT_HASH_FUNCTION dd_FNVM_hash_string
#elif ((DD_DEFAULT_HASH_ALGORITHM == DD_FNVM) && (DD_CASE_SENSITIVE_HASHING == false))
#define DD_DEFAULT_HASH_FUNCTION dd_FNVM_hash_string_toupper
#elif ((DD_DEFAULT_HASH_ALGORITHM == DD_HSIEH) && (DD_CASE_SENSITIVE_HASHING == true))
#define DD_DEFAULT_HASH_FUNCTION dd_HSIEH_hash_string
#elif ((DD_DEFAULT_HASH_ALGORITHM == DD_HSIEH) && (DD_CASE_SENSITIVE_HASHING == false))
#define DD_DEFAULT_HASH_FUNCTION dd_HSIEH_hash_string_toupper
#else
#define DD_DEFAULT_HASH_FUNCTION dd_SDBM_hash_string
#endif

// --------------------------------------------------------------------------
// Range check for DD_AUTOMATICALLY_TRIM_KEYS,  must be true or false
// --------------------------------------------------------------------------

#if ((DD_AUTOMATICALLY_TRIM_KEYS != true) && (DD_AUTOMATICALLY_TRIM_KEYS != false))
#undef DD_AUTOMATICALLY_TRIM_KEYS
#define DD_AUTOMATICALLY_TRIM_KEYS true
#endif

// --------------------------------------------------------------------------
// Range check for DD_PRINTABLE_7BIT_ASCII_KEYS_ONLY,  must be true or false
// --------------------------------------------------------------------------

#if ((DD_PRINTABLE_7BIT_ASCII_KEYS_ONLY != true) && (DD_PRINTABLE_7BIT_ASCII_KEYS_ONLY != false))
#undef DD_PRINTABLE_7BIT_ASCII_KEYS_ONLY
#define DD_PRINTABLE_7BIT_ASCII_KEYS_ONLY false
#endif

// --------------------------------------------------------------------------
// Range check for DD_MINIMUM_DEFAULT_HASHTABLE_SIZE,  must be > 0
// --------------------------------------------------------------------------

#if (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 0)
#undef DD_MINIMUM_DEFAULT_HASHTABLE_SIZE
#define DD_MINIMUM_DEFAULT_HASHTABLE_SIZE 256
#endif

// --------------------------------------------------------------------------
// Set DD_DEFAULT_HASHTABLE_SIZE,  default hashtable size
// --------------------------------------------------------------------------

#if (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 7)
#define DD_DEFAULT_HASHTABLE_SIZE   7
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 11)
#define DD_DEFAULT_HASHTABLE_SIZE   11
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 17)
#define DD_DEFAULT_HASHTABLE_SIZE   17
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 37)
#define DD_DEFAULT_HASHTABLE_SIZE   37
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 67)
#define DD_DEFAULT_HASHTABLE_SIZE   67
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 131)
#define DD_DEFAULT_HASHTABLE_SIZE   131
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 193)
#define DD_DEFAULT_HASHTABLE_SIZE   193
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 257)
#define DD_DEFAULT_HASHTABLE_SIZE   257
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 521)
#define DD_DEFAULT_HASHTABLE_SIZE   521
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 787)
#define DD_DEFAULT_HASHTABLE_SIZE   787
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 1031)
#define DD_DEFAULT_HASHTABLE_SIZE   1031
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 1283)
#define DD_DEFAULT_HASHTABLE_SIZE   1283
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 1543)
#define DD_DEFAULT_HASHTABLE_SIZE   1543
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 1801)
#define DD_DEFAULT_HASHTABLE_SIZE   1801
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 2053)
#define DD_DEFAULT_HASHTABLE_SIZE   2053
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 2309)
#define DD_DEFAULT_HASHTABLE_SIZE   2309
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 2579)
#define DD_DEFAULT_HASHTABLE_SIZE   2579
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 2819)
#define DD_DEFAULT_HASHTABLE_SIZE   2819
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 3079)
#define DD_DEFAULT_HASHTABLE_SIZE   3079
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 3329)
#define DD_DEFAULT_HASHTABLE_SIZE   3329
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 3593)
#define DD_DEFAULT_HASHTABLE_SIZE   3593
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 3847)
#define DD_DEFAULT_HASHTABLE_SIZE   3847
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 4099)
#define DD_DEFAULT_HASHTABLE_SIZE   4099
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 5147)
#define DD_DEFAULT_HASHTABLE_SIZE   5147
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 6421)
#define DD_DEFAULT_HASHTABLE_SIZE   6421
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 7681)
#define DD_DEFAULT_HASHTABLE_SIZE   7681
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 8963)
#define DD_DEFAULT_HASHTABLE_SIZE   8963
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 10243)
#define DD_DEFAULT_HASHTABLE_SIZE   10243
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 12809)
#define DD_DEFAULT_HASHTABLE_SIZE   12809
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 14593)
#define DD_DEFAULT_HASHTABLE_SIZE   14593
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 16411)
#define DD_DEFAULT_HASHTABLE_SIZE   16411
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 19207)
#define DD_DEFAULT_HASHTABLE_SIZE   19207
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 22027)
#define DD_DEFAULT_HASHTABLE_SIZE   22027
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 25097)
#define DD_DEFAULT_HASHTABLE_SIZE   25097
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 27917)
#define DD_DEFAULT_HASHTABLE_SIZE   27917
#elif (DD_MINIMUM_DEFAULT_HASHTABLE_SIZE <= 30211)
#define DD_DEFAULT_HASHTABLE_SIZE   30211
#else
#define DD_DEFAULT_HASHTABLE_SIZE   32771
#endif

// --------------------------------------------------------------------------
// Set DD_DEFAULT_SYMTABLE_SIZE,  default symbol table size
// --------------------------------------------------------------------------

#if (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 7)
#define DD_DEFAULT_SYMTABLE_SIZE   7
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 11)
#define DD_DEFAULT_SYMTABLE_SIZE   11
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 17)
#define DD_DEFAULT_SYMTABLE_SIZE   17
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 37)
#define DD_DEFAULT_SYMTABLE_SIZE   37
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 67)
#define DD_DEFAULT_SYMTABLE_SIZE   67
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 131)
#define DD_DEFAULT_SYMTABLE_SIZE   131
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 193)
#define DD_DEFAULT_SYMTABLE_SIZE   193
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 257)
#define DD_DEFAULT_SYMTABLE_SIZE   257
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 521)
#define DD_DEFAULT_SYMTABLE_SIZE   521
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 787)
#define DD_DEFAULT_SYMTABLE_SIZE   787
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 1031)
#define DD_DEFAULT_SYMTABLE_SIZE   1031
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 1283)
#define DD_DEFAULT_SYMTABLE_SIZE   1283
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 1543)
#define DD_DEFAULT_SYMTABLE_SIZE   1543
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 1801)
#define DD_DEFAULT_SYMTABLE_SIZE   1801
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 2053)
#define DD_DEFAULT_SYMTABLE_SIZE   2053
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 2309)
#define DD_DEFAULT_SYMTABLE_SIZE   2309
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 2579)
#define DD_DEFAULT_SYMTABLE_SIZE   2579
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 2819)
#define DD_DEFAULT_SYMTABLE_SIZE   2819
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 3079)
#define DD_DEFAULT_SYMTABLE_SIZE   3079
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 3329)
#define DD_DEFAULT_SYMTABLE_SIZE   3329
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 3593)
#define DD_DEFAULT_SYMTABLE_SIZE   3593
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 3847)
#define DD_DEFAULT_SYMTABLE_SIZE   3847
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 4099)
#define DD_DEFAULT_SYMTABLE_SIZE   4099
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 5147)
#define DD_DEFAULT_SYMTABLE_SIZE   5147
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 6421)
#define DD_DEFAULT_SYMTABLE_SIZE   6421
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 7681)
#define DD_DEFAULT_SYMTABLE_SIZE   7681
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 8963)
#define DD_DEFAULT_SYMTABLE_SIZE   8963
#elif (DD_MINIMUM_DEFAULT_SYMBOL_TABLE_SIZE <= 10243)
#define DD_DEFAULT_SYMTABLE_SIZE   10243
#else
#define DD_DEFAULT_SYMTABLE_SIZE   10243
#endif

// --------------------------------------------------------------------------
// check DD_USE_BOEHM_GC, must be true or false
// --------------------------------------------------------------------------
// defines macros _ALLOCATE() and _DEALLOCATE() accordingly

#if (DD_USE_BOEHM_GC == true)
#define _ALLOCATE(_size) GC_MALLOC(_size)
#define _DEALLOCATE(_ptr) GC_FREE(_ptr)
#else
#define _ALLOCATE(_size) malloc(_size)
#define _DEALLOCATE(_ptr) free(_ptr)
#endif

// --------------------------------------------------------------------------
// check DD_USE_PTHREADS, must be true or false
// --------------------------------------------------------------------------
// defines macros _MUTEX(), _INIT_MUTEX(), _LOCK(), _UNLOCK() and
// _DESTROY_MUTEX() accordingly

#if (DD_USE_PTHREADS == true)
#define _MUTEX(_mutex) pthread_mutex_t _mutex
#define _INIT_MUTEX(_mutex) pthread_mutex_init(&_mutex, NULL)
#define _LOCK(_mutex) pthread_mutex_lock(&_mutex)
#define _UNLOCK(_mutex) pthread_mutex_unlock(&_mutex)
#define _DESTROY_MUTEX(_mutex) pthread_mutex_destroy(&_mutex)
#else
#define _MUTEX(_mvar) /* pthread_mutex_t _mutex; */
#define _INIT_MUTEX(_mutex) /* pthread_mutex_init(&_mutex, NULL) */
#define _LOCK(_mutex) /* pthread_mutex_lock(&_mutex) */
#define _UNLOCK(_mutex) /* pthread_mutex_unlock(&_mutex) */
#define _DESTROY_MUTEX(_mutex) /* pthread_mutex_destroy(&_mutex) */
#endif

// --------------------------------------------------------------------------
// generic attribute macros
// --------------------------------------------------------------------------

#ifdef _ATTR
#undef _ATTR
#endif

#define _nn_ nonnull
#define align2 aligned(2)
#define align4 aligned(4)
#define align8 aligned(8)
#define _ATTR(_attr, ...) __attribute__((_attr))

// --------------------------------------------------------------------------
// function attribute: always inline private functions
// --------------------------------------------------------------------------

#ifdef _ai_
#undef _ai_
#endif

#if (DD_ALWAYS_INLINE_PRIVATE_FUNCTIONS == true)
#define _ai_ always_inline
#else
#define _ai_ auto_inline /* this will cause warning "directive ignored" */
#endif


#endif

// END OF FILE